#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/error.h>
#include <nemo/pathexec.h>
#include <nemo/prot.h>
#include <nemo/exit.h>
#include <nemo/unix.h>
#include <nemo/str.h>

extern char *crypt(const char *key, const char *salt);

#include <pwd.h>
static struct passwd *pw;

#include "hasspnam.h"
#ifdef HASGETSPNAM
#include <shadow.h>
static struct spwd *spw;
#endif

#include "hasuserpw.h"
#ifdef HASUSERPW
#include <userpw.h>
static struct userpw *upw;
#endif

static char up[513];
static size_t uplen;

int main(int argc, char **argv)
{
  char *login;
  char *password;
  char *encrypted;
  char *stored;
  unsigned int i;
  ssize_t r;

  if (argc < 2) _exit(2);

  stored = 0;
  uplen = 0;
  for (;;) {
    do {
      r = read(3, up + uplen, sizeof(up) - uplen);
    } while ((r < 0) && (errno == error_intr));
    if (r < 0) _exit(111);
    if (!r) break;
    uplen += (size_t)r;
    if (uplen >= sizeof(up)) _exit(1);
  }
  close(3);

  i = 0;
  if (i >= uplen) _exit(2);
  login = up + i;
  while (up[i++]) {
    if (i >= uplen) _exit(2);
  }
  password = up + i;
  if (i >= uplen) _exit(2);
  while (up[i++]) {
    if (i >= uplen) _exit(2);
  }

  pw = getpwnam(login);
  if (pw) {
    stored = pw->pw_passwd;
  }
  else {
    if (errno == error_txtbsy) _exit(111);
    _exit(1);
  }
#ifdef HASUSERPW
  upw = getuserpw(login);
  if (upw) {
    stored = upw->upw_passwd;
  }
  else {
    if (errno == error_txtbsy) _exit(111);
  }
#endif
#ifdef HASGETSPNAM
  spw = getspnam(login);
  if (spw) {
    stored = spw->sp_pwdp;
  }
  else {
    if (errno == error_txtbsy) _exit(111);
  }
#endif
  if (!stored) _exit(1);

  encrypted = crypt(password, stored);
  for (i = 0; i < sizeof(up); ++i) {
    up[i] = 0;
  }

  if (!*stored || str_diff(encrypted, stored)) _exit(1);

  if (prot_gid(pw->pw_gid) < 0) _exit(1);
  if (prot_uid(pw->pw_uid) < 0) _exit(1);
  if (chdir(pw->pw_dir) < 0) _exit(111);

  if (!pathexec_env("USER", pw->pw_name)) _exit(111);
  if (!pathexec_env("HOME", pw->pw_dir)) _exit(111);
  if (!pathexec_env("SHELL", pw->pw_shell)) _exit(111);
  pathexec(argv + 1);
  _exit(111);
}
